#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import argparse

import c4.cmany as cmany
from c4.cmany import util, args as c4args
from c4.cmany import variant
from c4.cmany import build_item


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test00VariantSpec(ut.TestCase):

    def c(self, name, var, **ref):
        self.assertEqual(var.name, name)
        for k, refval in ref.items():
            result = getattr(var.flags, k)
            self.assertEqual(result, refval)

    def test00_simple(self):
        v = variant.Variant.create_variants("somevariant: -X '-fPIC'")
        v = v[0]
        self.c('somevariant', v, cmake_vars=[], defines=[], cflags=[], cxxflags=['-fPIC'])

    def test01_nospecs(self):
        vars = ['var0','var1','var2']
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 3)
            self.c('var0', out[0], cmake_vars=[], defines=[], cflags=[], cxxflags=[])
            self.c('var1', out[1], cmake_vars=[], defines=[], cflags=[], cxxflags=[])
            self.c('var2', out[2], cmake_vars=[], defines=[], cflags=[], cxxflags=[])

    def test02_norefs(self):
        vars = ['\'var0: -X "-fPIC" -D VAR1,VAR_TYPE=1\'',
                '\'var1: -X "-Wall" -D VAR2,VAR_TYPE=2\'',
                '\'var2: -X nortti,c++14 -D VAR3,VAR_TYPE=3\'',
        ]
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 3)
            self.c('var0', out[0], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-fPIC'])
            self.c('var1', out[1], cmake_vars=[], defines=['VAR2', 'VAR_TYPE=2'], cflags=[], cxxflags=['-Wall'])
            self.c('var2', out[2], cmake_vars=[], defines=['VAR3', 'VAR_TYPE=3'], cflags=[], cxxflags=['nortti', 'c++14'])

    def test10_refs_beginning(self):
        vars = [
            '\'var0: -X "-fPIC" -D VAR1,VAR_TYPE=1\'',
            '\'var1: @var0 -X "-Wall" -D VAR2,VAR_TYPE=2\'',
            '\'var2: @var1 -X nortti,c++14 -D VAR3,VAR_TYPE=3\'',
        ]
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 3)
            self.c('var0', out[0], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-fPIC'])
            self.c('var1', out[1], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2'], cflags=[], cxxflags=['-fPIC', '-Wall'])
            self.c('var2', out[2], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2', 'VAR3', 'VAR_TYPE=3'], cflags=[], cxxflags=['-fPIC', '-Wall', 'nortti', 'c++14'])

    def test11_refs_middle(self):
        vars = [
            '\'var0: -X "-fPIC" -D VAR1,VAR_TYPE=1\'',
            '\'var1: -X "-Wall" @var0 -D VAR2,VAR_TYPE=2\'',
            '\'var2: -X nortti,c++14 @var1 -D VAR3,VAR_TYPE=3\'',
        ]
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 3)
            self.c('var0', out[0], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-fPIC'])
            self.c('var1', out[1], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2'], cflags=[], cxxflags=['-Wall', '-fPIC'])
            self.c('var2', out[2], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2', 'VAR3', 'VAR_TYPE=3'], cflags=[], cxxflags=['nortti', 'c++14', '-Wall', '-fPIC'])

    def test12_refs_end(self):
        vars = [
            '\'var0: -X "-fPIC" -D VAR1,VAR_TYPE=1\'',
            '\'var1: -X "-Wall" -D VAR2,VAR_TYPE=2 @var0\'',
            '\'var2: -X nortti,c++14 -D VAR3,VAR_TYPE=3 @var1\'',
        ]
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 3)
            self.c('var0', out[0], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-fPIC'])
            self.c('var1', out[1], cmake_vars=[], defines=['VAR2', 'VAR_TYPE=2', 'VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-Wall', '-fPIC'])
            self.c('var2', out[2], cmake_vars=[], defines=['VAR3', 'VAR_TYPE=3', 'VAR2', 'VAR_TYPE=2', 'VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['nortti', 'c++14', '-Wall', '-fPIC'])

    def test22_consecutive_refs_only(self):
        vars = [
            '\'var0: -X "-fPIC" -D VAR1,VAR_TYPE=1\'',
            '\'var1: -X "-Wall" -D VAR2,VAR_TYPE=2\'',
            '\'var2: -X "-g3" -D VAR3,VAR_TYPE=3\'',
            '\'var3: @var0 @var1\'',
            '\'var4: @var0 @var1 @var2\'',
        ]
        for w in (vars, ','.join(vars)):
            out = variant.Variant.create_variants(w)
            self.assertEquals(len(out), 5)
            self.c('var0', out[0], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1'], cflags=[], cxxflags=['-fPIC'])
            self.c('var1', out[1], cmake_vars=[], defines=['VAR2', 'VAR_TYPE=2'], cflags=[], cxxflags=['-Wall'])
            self.c('var2', out[2], cmake_vars=[], defines=['VAR3', 'VAR_TYPE=3'], cflags=[], cxxflags=['-g3'])
            self.c('var3', out[3], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2'], cflags=[], cxxflags=['-fPIC', '-Wall'])
            self.c('var4', out[4], cmake_vars=[], defines=['VAR1', 'VAR_TYPE=1', 'VAR2', 'VAR_TYPE=2', 'VAR3', 'VAR_TYPE=3'], cflags=[], cxxflags=['-fPIC', '-Wall', '-g3'])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test10AsArguments(ut.TestCase):

    def t(self, input, expected):
        parser = argparse.ArgumentParser()
        c4args.add_select(parser)
        args = parser.parse_args(input)
        with self.subTest(input=input, desc='argument parsing'):
            self.assertEqual(args.variants, expected, msg=input)

    def tp(self, input, which_var, expected_name, **expected_props):
        parser = argparse.ArgumentParser()
        c4args.add_select(parser)
        args = parser.parse_args(input)
        vars = variant.Variant.create_variants(args.variants)
        self.assertTrue(which_var < len(vars))
        var = vars[which_var]
        self.assertEqual(var.name, expected_name)
        for k, refval in expected_props.items():
            result = getattr(var.flags, k)
            with self.subTest(input=input, which_var=which_var, desc='variant properties', prop=k):
                self.assertEqual(result, refval, msg='property: ' + k)

    def r(self, input, expected):
        isl = list(util.intersperse_l("-v", input))
        self.t(isl, expected)
        isl = list(util.intersperse_l("--variant", input))
        self.t(isl, expected)

    def rp(self, input, which_var, expected_name, **expected_props):
        isl = list(util.intersperse_l("-v", input))
        self.tp(isl, which_var, expected_name, **expected_props)
        isl = list(util.intersperse_l("--variant", input))
        self.tp(isl, which_var, expected_name, **expected_props)

    def test00(self):
        self.t([], ['none'])
        self.t(["-v", "none"], ['none'])

    def test01(self):
        self.t(["-v", "none", "-v", 'xdebug: --cxxflags g3'], ['none', 'xdebug: --cxxflags g3'])
        self.t(["-v", "none,'xdebug: --cxxflags g3'"], ['none', 'xdebug: --cxxflags g3'])

    def test02(self):
        vars = [
            "none",
            "just_no_rtti: --cxxflags no_rtti",
            "fast: @just_no_rtti --cxxflags no_exceptions,no_bufsec,no_iterator_debug,fast_math"
        ]
        #
        self.r(vars, vars)
        # join first two
        j = '{},"{}"'.format(vars[0], vars[1])
        self.r([j, vars[2]], vars)
        # join second two
        j = '"{}","{}"'.format(vars[1], vars[2])
        self.r([vars[0], j], vars)
        # join all
        j = '"{}","{}","{}"'.format(vars[0], vars[1], vars[2])
        self.r([j], vars)

    def test03(self):
        vars = [
            "none",
            "dbg3: -X '-g3' -D DBG3",
            "wall: -X '-Wall' -D WALL",
            "dbg3_wall: @dbg3 @wall"
        ]
        def comp(input):
            with self.subTest(input=input):
                self.r(input, vars)
                self.rp(input, 0, 'none')
                self.rp(input, 1, 'dbg3', cxxflags=['-g3'], defines=['DBG3'])
                self.rp(input, 2, 'wall', cxxflags=['-Wall'], defines=['WALL'])
                self.rp(input, 3, 'dbg3_wall', cxxflags=['-g3','-Wall'], defines=['DBG3','WALL'])
        # all separate
        comp(vars)
        # join first two
        j = '{},"{}"'.format(vars[0], vars[1])
        comp([j, vars[2], vars[3]])
        # join second two
        j = '"{}","{}"'.format(vars[1], vars[2])
        comp([vars[0], j, vars[3]])
        # join third two
        j = '"{}","{}"'.format(vars[2], vars[3])
        comp([vars[0], vars[1], j])
        # join first three
        j = '"{}","{}","{}"'.format(vars[0], vars[1], vars[2])
        comp([j, vars[3]])
        # join all
        j = '"{}","{}","{}","{}"'.format(vars[0], vars[1], vars[2], vars[3])
        comp([j])

    def test04(self):
        s = ['-v', "none,\"just_no_rtti: --cxxflags no_rtti\"",
             '-v', "fast: @just_no_rtti --cxxflags no_exceptions,no_bufsec,no_iterator_debug,fast_math"]
        self.tp(s, 0, 'none')
        self.tp(s, 1, 'just_no_rtti')
        self.tp(s, 2, 'fast')

    def test05_nospecs(self):
        vars = ["var0", "var1", "var2", "var3"]
        def comp(input):
            with self.subTest(input=input):
                self.r(input, vars)
                self.rp(input, 0, 'var0', cmake_vars=[], cflags=[], cxxflags=[], defines=[])
                self.rp(input, 1, 'var1', cmake_vars=[], cflags=[], cxxflags=[], defines=[])
                self.rp(input, 2, 'var2', cmake_vars=[], cflags=[], cxxflags=[], defines=[])
                self.rp(input, 3, 'var3', cmake_vars=[], cflags=[], cxxflags=[], defines=[])
        # all separate
        comp(vars)
        for q in ('"', "'"):
            # join first two
            j = '{},{quote}{}{quote}'.format(vars[0], vars[1], quote=q)
            comp([j, vars[2], vars[3]])
            # join second two
            j = '{quote}{}{quote},{quote}{}{quote}'.format(vars[1], vars[2], quote=q)
            comp([vars[0], j, vars[3]])
            # join third two
            j = '{quote}{}{quote},{quote}{}{quote}'.format(vars[2], vars[3], quote=q)
            comp([vars[0], vars[1], j])
            # join first three
            j = '{quote}{}{quote},{quote}{}{quote},{quote}{}{quote}'.format(vars[0], vars[1], vars[2], quote=q)
            comp([j, vars[3]])
            # join all
            j = '{quote}{}{quote},{quote}{}{quote},{quote}{}{quote},{quote}{}{quote}'.format(vars[0], vars[1], vars[2], vars[3], quote=q)
            comp([j])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test11AsUnquotedArguments(ut.TestCase):
    """sometimes (when?) we received the arguments with even interior quotes
    removed. So variant parsing must be able to deal with fully unquoted specs.
    """

    def t(self, input, expected):
        variants = build_item.BuildItem.parse_args(input)
        with self.subTest(input=input):
            self.assertEqual(variants, expected, msg=input)

    def test00(self):
        var0 = 'none'
        var1 = 'just_no_rtti: --cxxflags no_rtti'
        self.t(var0 + ',' + var1, [var0, var1])

    def test01(self):
        var0 = 'none'
        var1 = 'just_no_rtti: --cxxflags no_rtti'
        var2 = 'fast: @just_no_rtti --cxxflags no_exceptions,no_bufsec,no_iterator_debug,fast_math'
        self.t(var0 + ',' + var1 + ',' + var2, [var0, var1, var2])

    def test02(self):
        var0 = 'var0'
        var1 = 'var1'
        var2 = 'var2'
        self.t(var0 + ',' + var1 + ',' + var2, [var0, var1, var2])

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    ut.main()




import re
import os
import tempfile

from collections import OrderedDict as odict

from .conf import USER_DIR
from .util import cacheattr, setcwd, runsyscmd
from . import util

_cache_entry = r'^(.*?)(:.*?)=(.*)$'


def setcachevar(builddir, var, value):
    setcachevars(builddir, odict([(var, value)]))


def getcachevar(builddir, var):
    v = getcachevars(builddir, [var])
    return v[var]


def setcachevars(builddir, varvalues):
    with setcwd(builddir, silent=True):
        with open('CMakeCache.txt', 'r') as f:
            ilines = f.readlines()
            olines = []
        for l in ilines:
            for k, v in varvalues.items():
                if l.startswith(k + ':'):
                    n = re.sub(_cache_entry, r'\1\2=' + v, l)
                    l = n
            olines.append(l)
        with open('CMakeCache.txt', 'w') as f:
            f.writelines(olines)


def getcachevars(builddir, varlist):
    vlist = [v + ':' for v in varlist]
    values = odict()
    with setcwd(builddir, silent=True):
        with open('CMakeCache.txt') as f:
            for line in f:
                for v in vlist:
                    if line.startswith(v):
                        ls = line.strip()
                        vt = re.sub(_cache_entry, r'\1', ls)
                        values[vt] = re.sub(_cache_entry, r'\3', ls)
    return values


def loadvars(builddir):
    """if builddir does not exist or does not have a cache, returns an
    empty odict"""
    v = odict()
    if builddir is None or not os.path.exists(builddir):
        return v
    c = os.path.join(builddir, 'CMakeCache.txt')
    if os.path.exists(c):
        with open(c, 'r') as f:
            for line in f:
                # print("loadvars0", line.strip())
                if not re.match(_cache_entry, line):
                    continue
                ls = line.strip()
                name = re.sub(_cache_entry, r'\1', ls)
                vartype = re.sub(_cache_entry, r'\2', ls)[1:]
                value = re.sub(_cache_entry, r'\3', ls)
                # print("loadvars1", name, vartype, value)
                v[name] = CMakeCache.Var(name, value, vartype)
    return v


# -----------------------------------------------------------------------------
class CMakeCache(odict):

    def __init__(self, builddir=None):
        super().__init__(loadvars(builddir))
        self.dirty = False

    def getvars(self, names):
        out = odict()
        for n in names:
            v = self.get(n)
            out[n] = v
        return out

    def b(self, name, val, **kwargs):
        """set a boolean"""
        return self.setvar(name, val, "BOOL", **kwargs)

    def s(self, name, val, **kwargs):
        """set a string"""
        return self.setvar(name, val, "STRING", **kwargs)

    def p(self, name, val, **kwargs):
        """set a path to a dir"""
        if util.in_windows():
            val = re.sub(r'\\', r'/', val)
        return self.setvar(name, val, "PATH", **kwargs)

    def f(self, name, val, **kwargs):
        """set a path to a file"""
        if util.in_windows():
            val = re.sub(r'\\', r'/', val)
        return self.setvar(name, val, "FILEPATH", **kwargs)

    def i(self, name, val, **kwargs):
        """set a cmake internal var"""
        return self.setvar(name, val, "INTERNAL", **kwargs)

    def setvar(self, name, val, vartype=None, **kwargs):
        v = self.get(name)
        if v is not None:
            changed = v.reset(val, vartype, **kwargs)
            self.dirty |= changed
            return changed
        else:
            v = CMakeCache.Var(name, val, vartype, dirty=True, **kwargs)
            self[name] = v
            self.dirty = True
            return True

    def commit(self, builddir):
        if (not self.dirty
            or builddir is None
            or not os.path.exists(builddir)
            or not os.path.exists(os.path.join(builddir, 'CMakeCache.txt'))):
            return False
        tmp = odict()
        for _, v in self.items():
            if not v.dirty:
                continue
            tmp[v.name] = v.val
        setcachevars(builddir, tmp)
        for _, v in self.items():
            v.dirty = False
        self.dirty = False
        return True

    # -------------------------------------------------------------------------
    class Var:

        def __init__(self, name, val, vartype=None, dirty=False, from_input=False):
            self.name = name
            self.val = val
            self.vartype = 'STRING' if vartype is None else vartype
            self.dirty = dirty
            self.from_input = from_input

        def reset(self, val, vartype='', **kwargs):
            """

            :param val:
            :param vartype:
            :param kwargs:
                force_dirty, defaults to False
                from_input, defaults to None
            :return:
            """
            force_dirty = kwargs.get('force_dirty', False)
            from_input = kwargs.get('from_input')
            if from_input is not None:
                self.from_input = from_input
            if vartype == 'STRING' or (vartype is None and self.vartype == 'STRING'):
                candidates = (val, val.strip("'"), val.strip('"'))
                equal = False
                for c in candidates:
                    if c == self.val:
                        equal = True
                        break
            else:
                equal = (self.val == val)
            if not equal or (vartype is not None and vartype != self.vartype):
                self.val = val
                self.vartype = vartype if vartype is not None else self.vartype
                self.dirty = True
                return True
            if force_dirty:
                self.dirty = True
            return force_dirty

        def __repr__(self):
            return self.name + ':' + self.vartype + '=' + self.val

        def __str__(self):
            return self.name + ':' + self.vartype + '=' + self.val


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class CMakeSysInfo:
    """encapsulates the results returned from
    `cmake [-G <which_generator>][-T <toolset>] --system-information`.
    This is used for selecting default values for system, compiler,
    generator, etc."""

    @staticmethod
    def generator():
        return cacheattr(__class__, '_generator_default',
                         lambda: __class__._getstr('CMAKE_GENERATOR', 'default'))

    @staticmethod
    def system_name(which_generator="default"):
        return __class__.var('CMAKE_SYSTEM_NAME', which_generator, lambda v: v.lower())

    @staticmethod
    def architecture(which_generator="default"):
        return __class__.var('CMAKE_SYSTEM_PROCESSOR', which_generator, lambda v: v.lower())

    @staticmethod
    def cxx_compiler(which_generator="default"):
        return __class__.var('CMAKE_CXX_COMPILER', which_generator)

    @staticmethod
    def c_compiler(which_generator="default"):
        return __class__.var('CMAKE_C_COMPILER', which_generator)

    @staticmethod
    def var(var_name, which_generator="default", transform_fn=lambda x: x):
        gs = __class__._getstr
        return cacheattr(__class__, '_{}_{}'.format(var_name, _genid(which_generator)),
                         lambda: transform_fn(gs(var_name, which_generator)))

    @staticmethod
    def info(which_generator="default"):
        return cacheattr(__class__, '_info_' + _genid(which_generator),
                         lambda: __class__.system_info(which_generator))

    @staticmethod
    def _getstr(var_name, which_generator):
        regex = r'^' + var_name + r' "(.*)"'
        for l in __class__.info(which_generator):
            if l.startswith(var_name):
                l = l.strip("\n").lstrip(" ").rstrip(" ")
                # print(var_name, "startswith :", l)
                if re.match(regex, l):
                    s = re.sub(regex, r'\1', l)
                    # print(var_name, "result: '" + s + "'")
                    return s
        err = "could not find variable {} in the output of `cmake --system-information` for generator {}"
        err = err.format(var_name, which_generator)
        raise Exception(err.format(var_name))

    @staticmethod
    def system_info(gen):
        """generator can be a string or a cmany.Generator object"""
        from .generator import Generator
        # print("CMakeSystemInfo: asked info for", which_generator)
        p = _genid(gen)
        d = os.path.join(USER_DIR, 'cmake_info', p)
        p = os.path.join(d, 'info')
        if os.path.exists(p):
            # print("CMakeSystemInfo: asked info for", which_generator, "... found", p)
            with open(p, "r") as f:
                i = f.readlines()
        else:
            if isinstance(gen, Generator):
                cmd = ['cmake'] + gen.configure_args() + ['--system-information']
            else:
                if gen == "default" or gen == "":
                    cmd = ['cmake', '--system-information']
                else:
                    if gen.startswith('vs') or gen.startswith('Visual Studio'):
                        from . import vsinfo
                        gen = vsinfo.to_gen(gen)
                    cmd = ['cmake', '-G', str(gen), '--system-information']

            if not os.path.exists(d):
                os.makedirs(d)
            print("\ncmany: CMake information for generator '{}' was not found. Creating and storing...".format(gen))
            with setcwd(d):
                out = runsyscmd(cmd, echo_output=False, capture_output=True)
            print("cmany: finished generating information for generator '{}'\n".format(gen))
            with open(p, "w") as f:
                f.write(out)
            i = out.split("\n")
        return i


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def _genid(gen):
    from .generator import Generator
    p = gen.sysinfo_name if isinstance(gen, Generator) else gen
    p = re.sub(r'[() ]', '_', p)
    return p


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# def get_toolchain_cache(toolchain):
#     d = os.path.join(USER_DIR, 'toolchains', re.sub(os.sep, '+', toolchain))
#     print("toolchain cache: USER_DIR=", USER_DIR)
#     print("toolchain cache: d=", d)
#     bd = os.path.join(d, 'build')
#     print("toolchain cache: bd=", bd)
#     if not os.path.exists(d):
#         os.makedirs(d)
#         with setcwd(d):
#             with open('main.cpp', 'w') as f:
#                 f.write("int main() {}")
#             with open('CMakeLists.txt', 'w') as f:
#                 f.write("""
# cmake_minimum_required(VERSION 2.6)
# project(toolchain_test)
# add_executable(main main.cpp)
# """)
#         if not os.path.exists(bd):
#             os.makedirs(bd)
#         with setcwd(bd):
#             cmd = ['cmake', '-DCMAKE_TOOLCHAIN_FILE='+toolchain, '..']
#             runsyscmd(cmd, echo_output=True)
#     return loadvars(bd)
def extract_toolchain_compilers(toolchain):
    with open(toolchain) as f:
        lines = f.readlines()
        out = odict()
        for l in lines:
            res = re.search(r'(set|SET)\ ?\(\ ?(CMAKE_.*?_COMPILER) (.*?)\ ?\)', l)
            if res:
                res = res.groups()
                out[res[1]] = res[2]
        return out



#!/usr/bin/env python3

import unittest as ut
import subtest_fix
import os
import sys
import glob
import argparse
import copy
from itertools import combinations

import c4.cmany as cmany
import c4.cmany.util as util
import c4.cmany.main as main
import c4.cmany.cmake as cmake

from multiprocessing import cpu_count as cpu_count

srcdir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'src')
sys.path.insert(0, srcdir)
maincmd = [sys.executable, '-m', 'c4.cmany.main', '--show-args']

projdir = os.path.dirname(__file__)
compiler_set = os.environ.get('CMANY_TEST_COMPILERS', None)
build_types = os.environ.get('CMANY_TEST_BUILDTYPES', 'Debug,Release')
test_projs = os.environ.get('CMANY_TEST_PROJS', 'hello,libhello')
proj_targets = {
    'hello': {
        'lib': [],
        'exe': ['hello'],
    },
    'libhello': {
        'lib': ['hello', 'hello_static'],
        'exe': ['test_hello', 'test_hello_static'],
    },
}
flag_bundle_set = {
    'none': {
        'spec': 'none',
        'expected': {
            'none': {'vars': [], 'defines': [], 'cxxflags': [], 'flags': [], },
        },
    },
    'foo': {
        'spec': '\'foo: -V FOO_VAR=1 -D FOO_DEF=1 -X "wall" -C "wall"\'',
        'expected': {
            'foo': {'vars': ['FOO_VAR=1'], 'defines': ['FOO_DEF=1'], 'cxxflags': ['wall'], 'flags': ['wall'], },
        },
    },
    'bar': {
        'spec': '\'bar: -V BAR_VAR=1 -D BAR_DEF=1 -X "g3" -C "g3"\'',
        'expected': {
            'bar': {'vars': ['BAR_VAR=1'], 'defines': ['BAR_DEF=1'], 'cxxflags': ['g3'], 'flags': ['g3'], },
        },
    },
}
variant_set = [flag_bundle_set[v]['spec'] for v in ('none', 'foo', 'bar')]

variant_tests = {
    'variant_test00-null':[],
    'variant_test01-none_explicit':['none'],

    'variant_test10-foo_only':['foo'],
    'variant_test11-none_foo':['none', 'foo'],

    'variant_test20-bar_only':['bar'],
    'variant_test21-none_bar':['none', 'bar'],

    'variant_test30-foobar_only':['foo', 'bar'],
    'variant_test31-foobar_only':['none', 'foo', 'bar'],
}


def _get_variant_spec(test_name):
    blueprint = variant_tests[test_name]
    if not blueprint:
        return []
    li = ['-v'] + [','.join(flag_bundle_set[v]['spec']) for v in blueprint]
    variants = cmany.Variant.create_variants(li)
    return li, variants


# unset environment variables which affect the behaviour of child invokations
# of cmany
os.environ['CMANY_ARGS'] = ''
os.environ['CMANY_PFX_ARGS'] = ''


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class CMakeTestProj:

    def __init__(self, proj):
        self.proj = proj
        self.root = util.chkf(projdir, proj)
        if proj_targets.get(proj) is None:
            raise Exception("no target info for project " + proj)
        self.libs = proj_targets[proj]['lib']
        self.exes = proj_targets[proj]['exe']
        self.targets = self.libs + self.exes
        self.multi_target = (len(self.targets) > 1)
        # http://stackoverflow.com/questions/17176887/python-get-all-permutation-of-a-list-w-o-repetitions
        self.target_combinations = []
        for i in range(1, len(self.targets) + 1):
            self.target_combinations += list(combinations(self.targets, i))

    def run(self, args_, custom_root=None):
        args = copy.deepcopy(args_)
        root = self.root
        if custom_root is not None:
            with util.setcwd(self.root):
                root = os.path.abspath(custom_root)
                if not os.path.exists(root):
                    os.makedirs(root)
                projdir = os.path.abspath('.')
                args.append(projdir)
        with util.setcwd(root):
            args = maincmd + args
            #print("----->run():", self.proj, "at", os.getcwd(), " ".join(args))
            util.runsyscmd(args)
            #print("----->finished run():", self.proj, "at", os.getcwd(), " ".join(args))


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# prepare inputs

test_projs = util.splitesc(test_projs, ',')
projs = [CMakeTestProj(p) for p in test_projs]

if compiler_set is None:
    compiler_set = [cmany.Compiler.default()]
else:
    compiler_set = [cmany.Compiler(c) for c in util.splitesc(compiler_set, ',')]

build_types = [cmany.BuildType(b) for b in util.splitesc(build_types, ',')]

variant_set = cmany.Variant.create_variants(variant_set)

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def run_projs(testobj, args, check_fn=None):
    numbuilds = len(compiler_set) * len(build_types) * len(variant_set)

    # run with default parameters
    bd = '.test/0--default--build'
    id = '.test/0--default--install'
    for p in projs:
        with testobj.subTest(msg="default parameters", proj=p.proj):
            p.run(args + ['--build-dir', bd, '--install-dir', id])
            if check_fn:
                tb = TestBuild(proj=p, buildroot=bd, installroot=id,
                               compiler=cmany.Compiler.default(),
                               build_type=cmany.BuildType.default(),
                               variant=cmany.Variant.default(),
                               numbuilds=1)
                check_fn(tb)

    # run with default parameters in a non root dir
    rd = '.test/1--non_root_dir'
    for p in projs:
        with testobj.subTest(msg="run in a non root dir", proj=p.proj):
            p.run(args, custom_root=rd)
            if check_fn:
                tb = TestBuild(proj=p, buildroot=bd, installroot=id,
                               compiler=cmany.Compiler.default(),
                               build_type=cmany.BuildType.default(),
                               variant=cmany.Variant.default(),
                               numbuilds=1)
                check_fn(tb)

    if numbuilds == 1:
        return

    # run all sys,arch,compiler,buildtype,variant combinations at once
    bd = '.test/2--comps{}--types{}--variants{}--build'.format(len(compiler_set), len(build_types), len(variant_set))
    id = '.test/2--comps{}--types{}--variants{}--install'.format(len(compiler_set), len(build_types), len(variant_set))
    for p in projs:
        with testobj.subTest(msg="run all combinations at once", proj=p.proj):
            p.run(args + ['--build-dir', bd,
                          '--install-dir', id,
                          '-c', ','.join([c.name if c.is_msvc else c.path for c in compiler_set]),
                          '-t', ','.join([str(b) for b in build_types]),
                          '-v', ','.join([v.full_specs for v in variant_set])
            ])
            if check_fn:
                for c in compiler_set:
                    for t in build_types:
                        for v in variant_set:
                            tb = TestBuild(proj=p, buildroot=bd, installroot=id,
                                           compiler=c, build_type=t, variant=v,
                                           numbuilds=numbuilds)
                            check_fn(tb)

    # run sys,arch,compiler,buildtype combinations individually
    for p in projs:
        for c in compiler_set:
            for t in build_types:
                for v in variant_set:
                    with testobj.subTest(msg="run all combinations individually",
                                         proj=p.proj, compiler=c, build_type=t, variant=v):
                        bd = '.test/3--{}--{}--{}--build'.format(c, t, v.name)
                        id = '.test/3--{}--{}--{}--install'.format(c, t, v.name)
                        p.run(args + ['--build-dir', bd,
                                      '--install-dir', id,
                                      '-c', c.name if c.is_msvc else c.path,
                                      '-t', str(t),
                                      '-v', v.full_specs,
                        ])
                        if check_fn:
                            tb = TestBuild(proj=p, buildroot=bd, installroot=id,
                                           compiler=c, build_type=t, variant=v,
                                           numbuilds=1)
                            check_fn(tb)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class TestBuild:

    def __init__(self, proj, buildroot, installroot, compiler, build_type, variant, numbuilds):
        self.proj = proj
        self.buildroot = buildroot
        self.installroot = installroot
        self.compiler = compiler
        self.build_type = build_type
        self.variant = variant
        self.numbuilds = numbuilds
        self.flags = cmany.BuildFlags('all_builds')
        self.build_obj = cmany.Build(proj_root=self.proj.root,
                                     build_root=os.path.join(self.proj.root, self.buildroot),
                                     install_root=os.path.join(self.proj.root, self.installroot),
                                     system=cmany.System.default(),
                                     arch=cmany.Architecture.default(),
                                     build_type=build_type,
                                     compiler=compiler,
                                     variant=variant,
                                     flags=self.flags,
                                     num_jobs=cpu_count(),
                                     kwargs={}
                                     )

    def checkc(self, tester):
        tester.assertEqual(self.nsiblings(self.buildroot), self.numbuilds, msg=self.buildroot + str(self.siblings(self.buildroot)))
        build_type = cmake.getcachevar(self.build_obj.builddir, 'CMAKE_BUILD_TYPE')
        tester.assertEqual(build_type, str(self.build_type))

    def checkv(self, tester):
        pass

    def checkb(self, tester):
        self.checkc(tester)

    def checki(self, tester):
        tester.assertEqual(self.nsiblings(self.installroot), self.numbuilds)

    def nsiblings(self, dir):
        return len(self.siblings(dir))

    def siblings(self, dir):
        res = os.path.join(self.proj.root, dir, '*')
        ch = glob.glob(res)
        return ch

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test00Help(ut.TestCase):

    def test00_mh(self):
        projs[0].run(['-h'])

    def test01_mmhelp(self):
        projs[0].run(['--help'])

    def test02_cmd_mhelp(self):
        for c, aliases in main.cmds.items():
            if c == 'help':
                continue
            projs[0].run([c, '-h'])

    def test03_cmd_mmhelp(self):
        for c, aliases in main.cmds.items():
            if c == 'help':
                continue
            projs[0].run([c, '--help'])


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test01Configure(ut.TestCase):

    def test00_default(self):
        run_projs(self, ['c'], lambda tb: tb.checkc(self))

    def test01_custom_dirs(self):
        run_projs(self, ['c'], lambda tb: tb.checkc(self))


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test02Build(ut.TestCase):

    def test00_default(self):
        run_projs(self, ['b'], lambda tb: tb.checkb(self))

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class Test03Install(ut.TestCase):

    def test00_default(self):
        run_projs(self, ['i'], lambda tb: tb.checki(self))


class Test04Dependencies(ut.TestCase):
    pass

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    ut.main()



import os
import copy
import glob
import re
from datetime import datetime

from .generator import Generator
from . import util, cmake, vsinfo
from .named_item import NamedItem
from .variant import Variant
from .build_flags import BuildFlags
from .compiler import Compiler
from .architecture import Architecture

# experimental. I don't think it will stay unless conan starts accepting args
from .conan import Conan


# -----------------------------------------------------------------------------
class Build(NamedItem):
    """Holds a build's settings"""

    pfile = "cmany_preload.cmake"

    def __init__(self, proj_root, build_root, install_root,
                 system, arch, build_type, compiler, variant, flags,
                 num_jobs, kwargs):
        #
        self.kwargs = kwargs
        #
        self.projdir = util.chkf(proj_root)
        self.buildroot = os.path.abspath(build_root)
        self.installroot = os.path.abspath(install_root)
        #
        self.flags = flags
        self.system = system
        self.architecture = arch
        self.build_type = build_type
        self.compiler = compiler
        self.variant = variant

        self.adjusted = False

        if util.in_64bit and self.architecture.is32:
            if self.compiler.gcclike:
                self.compiler.make_32bit()
        elif util.in_32bit and self.architecture.is64:
            if self.compiler.gcclike:
                self.compiler.make_64bit()

        self._set_name_and_paths()  # calls super().__init__(self.tag)

        self.toolchain_file = self._get_toolchain()
        if self.toolchain_file:
            comps = cmake.extract_toolchain_compilers(self.toolchain_file)
            c = Compiler(comps['CMAKE_CXX_COMPILER'])
            self.adjust(compiler=c)

        # WATCHOUT: this may trigger a readjustment of this build's parameters
        self.generator = self.create_generator(num_jobs)

        # This will load the vars from the builddir cache, if it exists.
        # It should be done only after creating the generator.
        self.varcache = cmake.CMakeCache(self.builddir)
        # ... and this will overwrite (in memory) the vars with the input
        # arguments. This will make the cache dirty and so we know when it
        # needs to be committed back to CMakeCache.txt
        self.gather_input_cache_vars()

        self.deps = kwargs.get('deps', '')
        if self.deps and not os.path.isabs(self.deps):
            self.deps = os.path.abspath(self.deps)
        self.deps_prefix = kwargs.get('deps_prefix')
        if self.deps_prefix and not os.path.isabs(self.deps_prefix):
            self.deps_prefix = os.path.abspath(self.deps_prefix)
        if not self.deps_prefix:
            self.deps_prefix = self.builddir

    def _set_name_and_paths(self):
        self.tag = __class__.get_tag(
            self.system, self.architecture,
            self.compiler, self.build_type, self.variant, '-')
        super().__init__(self.tag)
        self.buildtag = self.tag
        self.installtag = self.tag  # this was different in the past and may become so in the future
        self.builddir = os.path.abspath(os.path.join(self.buildroot, self.buildtag))
        self.installdir = os.path.join(self.installroot, self.installtag)
        self.preload_file = os.path.join(self.builddir, Build.pfile)
        self.cachefile = os.path.join(self.builddir, 'CMakeCache.txt')

    def create_generator(self, num_jobs, fallback_generator="Unix Makefiles"):
        """create a generator, adjusting the build parameters if necessary"""
        #if self.toolchain_file is not None:
        #    toolchain_cache = cmake.get_toolchain_cache(self.toolchain_file)
        #    print(toolchain_cache)
        #    self.adjust(compiler=toolchain_cache['CMAKE_CXX_COMPILER'])
        if self.compiler.is_msvc:
            vsi = vsinfo.VisualStudioInfo(self.compiler.name)
            g = Generator(vsi.gen, self, num_jobs)
            arch = Architecture(vsi.architecture)
            self.adjust(architecture=arch)
            return g
        else:
            if self.system.name == "windows":
                return Generator(fallback_generator, self, num_jobs)
            else:
                return Generator(Generator.default_str(), self, num_jobs)

    def adjust(self, **kwargs):
        for k, _ in kwargs.items():
            if k not in ('architecture', 'compiler'):
                raise Exception("build adjustment for {} not supported".format(k))
        a = kwargs.get('architecture')
        if a and a != self.architecture:
            self.adjusted = True
            self.architecture = a
        c = kwargs.get('compiler')
        if c and c != self.compiler:
            self.adjusted = True
            self.compiler = c
        self._set_name_and_paths()

    @staticmethod
    def get_tag(s, a, c, t, v, sep='-'):
        # some utilities (eg, ar) dont deal well with + in the path
        # so replace + with x
        # eg see https://sourceforge.net/p/mingw/bugs/1429/
        sc = __class__.sanitize_compiler_name(c)
        s = str(s) + sep + str(a) + sep + sc + sep + str(t)
        if v is not None and isinstance(v, Variant):
            v = v.name
        if v and v != "none":
            s += "{sep}{var}".format(sep=sep, var=str(v))
        return s

    @staticmethod
    def sanitize_compiler_name(c):
        sc = re.sub(r'\+', 'x', str(c))
        return sc

    def create_dir(self):
        if not os.path.exists(self.builddir):
            os.makedirs(self.builddir)

    def configure_cmd(self, for_json=False):
        if for_json:
            return ('-C ' + self.preload_file
                    + ' ' + self.generator.configure_args(for_json))
        cmd = (['cmake', '-C', self.preload_file]
               + self.generator.configure_args())
        if self.kwargs.get('export_compile', False):
            cmd.append('-DCMAKE_EXPORT_COMPILE_COMMANDS=1')
        if self.toolchain_file:
            cmd.append('-DCMAKE_TOOLCHAIN_FILE=' + self.toolchain_file)
        cmd.append(self.projdir)
        return cmd

    def configure(self):
        self.create_dir()
        self.create_preload_file()
        self.handle_deps()
        if self.needs_cache_regeneration():
            self.varcache.commit(self.builddir)
        with util.setcwd(self.builddir, silent=False):
            cmd = self.configure_cmd()
            util.runsyscmd(cmd)
            self.mark_configure_done(cmd)

    def mark_configure_done(self, cmd):
        with util.setcwd(self.builddir):
            with open("cmany_configure.done", "w") as f:
                f.write(" ".join(cmd) + "\n")

    def needs_configure(self):
        if not os.path.exists(self.builddir):
            return True
        with util.setcwd(self.builddir):
            if not os.path.exists("cmany_configure.done"):
                return True
            if self.needs_cache_regeneration():
                return True
        return False

    def needs_cache_regeneration(self):
        if os.path.exists(self.cachefile) and self.varcache.dirty:
            return True
        return False

    def build(self, targets=[]):
        self.create_dir()
        with util.setcwd(self.builddir, silent=False):
            if self.needs_configure():
                self.configure()
            self.handle_deps()
            if len(targets) == 0:
                if self.compiler.is_msvc:
                    targets = ["ALL_BUILD"]
                else:
                    targets = ["all"]
            # cmake --build and visual studio won't handle
            # multiple targets at once, so loop over them.
            for t in targets:
                cmd = self.generator.cmd([t])
                util.runsyscmd(cmd)
            # this was written before using the loop above.
            # it can come to fail in some corner cases.
            self.mark_build_done(cmd)

    def mark_build_done(self, cmd):
        with util.setcwd(self.builddir):
            with open("cmany_build.done", "w") as f:
                f.write(" ".join(cmd) + "\n")

    def needs_build(self):
        if not os.path.exists(self.builddir):
            return True
        with util.setcwd(self.builddir):
            if not os.path.exists("cmany_build.done"):
                return True
            if self.needs_cache_regeneration():
                return True
        return False

    def install(self):
        self.create_dir()
        with util.setcwd(self.builddir, silent=False):
            if self.needs_build():
                self.build()
            cmd = self.generator.install()
            print(cmd)
            util.runsyscmd(cmd)

    def clean(self):
        self.create_dir()
        with util.setcwd(self.builddir):
            cmd = self.generator.cmd(['clean'])
            util.runsyscmd(cmd)
            os.remove("cmany_build.done")

    def _get_flagseq(self):
        return (
            self.flags,
            self.system.flags,
            self.architecture.flags,
            self.compiler.flags,
            self.build_type.flags,
            self.variant.flags
        )

    def _get_toolchain(self):
        tc = None
        for fs in self._get_flagseq():
            tc = BuildFlags.merge_toolchains(tc, fs.toolchain)
        if not tc:
            return None
        if not os.path.isabs(tc):
            tc = os.path.join(os.getcwd(), tc)
            tc = os.path.abspath(tc)
        if not os.path.exists(tc):
            raise Exception("file not found: " + tc)
        return tc

    def _gather_flags(self, which, append_to_sysinfo_var=None, with_defines=False):
        flags = []
        if append_to_sysinfo_var:
            try:
                flags = [cmake.CMakeSysInfo.var(append_to_sysinfo_var, self.generator)]
            except:
                pass
        # append overall build flags
        # append variant flags
        flagseq = self._get_flagseq()
        for fs in flagseq:
            wf = getattr(fs, which)
            for f in wf:
                if isinstance(f, str):
                    r = f
                elif isinstance(f, CFlag):
                    r = f.get(self.compiler)
                flags.append(r)
            if with_defines:
                flags += fs.defines
        # we're done
        return flags

    def _gather_cmake_vars(self):
        flagseq = self._get_flagseq()
        for fs in flagseq:
            for v in fs.cmake_vars:
                spl = v.split('=')
                vval = ''.join(spl[1:]) if len(spl) > 1 else ''
                nspl = spl[0].split(':')
                if len(nspl) == 1:
                    self.varcache.setvar(nspl[0], vval, from_input=True)
                elif len(nspl) == 2:
                    self.varcache.setvar(nspl[0], vval, nspl[1], from_input=True)
                else:
                    raise Exception('could not parse variable value: ' + v)

    def gather_input_cache_vars(self):
        self._gather_cmake_vars()
        vc = self.varcache
        #
        def _set(pfn, pname, pval): pfn(pname, pval, from_input=True)
        if (not self.generator.is_msvc) and (not self.toolchain_file):
            _set(vc.f, 'CMAKE_C_COMPILER', self.compiler.c_compiler)
            _set(vc.f, 'CMAKE_CXX_COMPILER', self.compiler.path)
        _set(vc.s, 'CMAKE_BUILD_TYPE', str(self.build_type))
        _set(vc.p, 'CMAKE_INSTALL_PREFIX', self.installdir)
        #
        cflags = self._gather_flags('cflags', 'CMAKE_C_FLAGS_INIT', with_defines=True)
        if cflags:
            _set(vc.s, 'CMAKE_C_FLAGS', ' '.join(cflags))
        #
        cxxflags = self._gather_flags('cxxflags', 'CMAKE_CXX_FLAGS_INIT', with_defines=True)
        if cxxflags:
            _set(vc.s, 'CMAKE_CXX_FLAGS', ' '.join(cxxflags))
        #
        # if self.flags.include_dirs:
        #     _set(vc.s, 'CMANY_INCLUDE_DIRECTORIES', ';'.join(self.flags.include_dirs))
        #
        # if self.flags.link_dirs:
        #     _set(vc.s, 'CMAKE_LINK_DIRECTORIES', ';'.join(self.flags.link_dirs))
        #

    def create_preload_file(self):
        # http://stackoverflow.com/questions/17597673/cmake-preload-script-for-cache
        self.create_dir()
        lines = []
        s = '_cmany_set({} "{}" {})'
        for _, v in self.varcache.items():
            if v.from_input:
                lines.append(s.format(v.name, v.val, v.vartype))
        if lines:
            tpl = _preload_file_tpl
        else:
            tpl = _preload_file_tpl_empty
        now = datetime.now().strftime("%Y/%m/%d %H:%m")
        txt = tpl.format(date=now, vars="\n".join(lines))
        with open(self.preload_file, "w") as f:
            f.write(txt)
        return self.preload_file

    @property
    def deps_done(self):
        dmark = os.path.join(self.builddir, "cmany_deps.done")
        exists = os.path.exists(dmark)
        return exists

    def mark_deps_done(self):
        with util.setcwd(self.builddir):
            with open("cmany_deps.done", "w") as f:
                s = ''
                if self.deps:
                    s += self.deps + '\n'
                if self.deps_prefix:
                    s += self.deps_prefix + '\n'
                f.write(s)

    def handle_deps(self):
        if self.deps_done:
            return
        if not self.deps:
            self.handle_conan()
            self.mark_deps_done()
            return
        util.lognotice(self.tag + ': building dependencies', self.deps)
        dup = copy.copy(self)
        dup.builddir = os.path.join(self.builddir, 'cmany_deps-build')
        dup.installdir = self.deps_prefix
        util.logwarn('installdir:', dup.installdir)
        dup.projdir = self.deps
        dup.preload_file = os.path.join(self.builddir, self.preload_file)
        dup.deps = None
        dup.generator.build = dup
        dup.configure()
        dup.build()
        try:
            # if the dependencies cmake project is purely consisted of
            # external projects, there won't be an install target.
            dup.install()
        except:
            util.logwarn(self.name + ": could not install. Maybe there's no install target?")
            pass
        util.logdone(self.name + ': finished building dependencies. Install dir=', self.installdir)
        self.varcache.p('CMAKE_PREFIX_PATH', self.installdir)
        self.mark_deps_done()

    def handle_conan(self):
        if not self.kwargs.get('with_conan'):
            return
        doit = False
        f = None
        for fn in ('conanfile.py', 'conanfile.txt'):
            f = os.path.join(self.projdir, fn)
            cf = os.path.join(self.builddir, 'conanbuildinfo.cmake')
            if os.path.exists(f) and not os.path.exists(cf):
                doit = True
                break
        if not doit:
            return
        util.logdone('found conan file')
        c = Conan()
        c.install(self)

    def json_data(self):
        """
        https://blogs.msdn.microsoft.com/vcblog/2016/11/16/cmake-support-in-visual-studio-the-visual-studio-2017-rc-update/
        https://blogs.msdn.microsoft.com/vcblog/2016/12/20/cmake-support-in-visual-studio-2017-whats-new-in-the-rc-update/
        """
        builddir = self.builddir.replace(self.projdir, '${projectDir}')
        builddir = re.sub(r'\\', r'/', builddir)
        return odict([
            ('name', self.tag),
            ('generator', self.generator.name),
            ('configurationType', self.build_type.name),
            ('buildRoot', builddir),
            ('cmakeCommandArgs', self.configure_cmd(for_json=True)),
            # ('variables', []),  # this is not needed since the vars are set in the preload file
        ])

    def get_targets(self):
        with util.setcwd(self.builddir):
            if self.generator.is_msvc:
                # each target in MSVC has a corresponding vcxproj file
                files = glob.glob(".", "*.vcxproj")
                files = [os.path.basename(f) for f in files]
                files = [os.path.splitext(f)[0] for f in files]
                return files
            elif self.generator.is_makefile:
                output = util.runsyscmd(["make", "help"], echo_cmd=False,
                                        echo_output=False, capture_output=True)
                output = output.split("\n")
                output = output[1:]  # The following are some of the valid targets....
                output = [o[4:] for o in output]  # take off the initial "... "
                output = [re.sub(r'(.*)\ \(the default if no target.*\)', r'\1', o) for o in output]
                output = sorted(output)
                result = []
                for o in output:
                    if o:
                        result.append(o)
                return result
            else:
                util.logerr("sorry, feature not implemented for this generator: " +
                            str(self.generator))

    def show_properties(self):
        util.logcmd(self.name)
        p = lambda n, v: print("{}={}".format(n, v))
        if self.toolchain_file:
            p('CMAKE_TOOLCHAIN_FILE', self.toolchain_file)
        p('CMAKE_C_COMPILER', self.compiler.c_compiler)
        p('CMAKE_CXX_COMPILER', self.compiler.path)
        dont_show = ('CMAKE_INSTALL_PREFIX', 'CMAKE_CXX_COMPILER', 'CMAKE_C_COMPILER')
        for _, v in self.varcache.items():
            if v.from_input:
                if v.name in dont_show:
                    continue
                p(v.name, v.val)
        p("PROJECT_BINARY_DIR", self.builddir)
        p("CMAKE_INSTALL_PREFIX", self.installdir)


# -----------------------------------------------------------------------------
_preload_file_tpl = ("""\
# Do not edit. Will be overwritten.
# Generated by cmany on {date}

if(NOT _cmany_set_def)
    set(_cmany_set_def ON)
    function(_cmany_set var value type)
        set(${{var}} "${{value}}" CACHE ${{type}} "")
        message(STATUS "cmany: ${{var}}=${{value}}")
    endfunction(_cmany_set)
endif(NOT _cmany_set_def)

message(STATUS "cmany:preload----------------------")
{vars}
message(STATUS "cmany:preload----------------------")

# if(CMANY_INCLUDE_DIRECTORIES)
#     include_directories(${{CMANY_INCLUDE_DIRECTORIES}})
# endif()
#
# if(CMANY_LINK_DIRECTORIES)
#     link_directories(${{CMANY_LINK_DIRECTORIES}})
# endif()

# Do not edit. Will be overwritten.
# Generated by cmany on {date}
""")

# -----------------------------------------------------------------------------
_preload_file_tpl_empty = ("""\
# Do not edit. Will be overwritten.
# Generated by cmany on {date}

message(STATUS "cmany: nothing to preload...")
""")



#!/usr/bin/env python3

import os
import re
import glob
import json
import sys

from collections import OrderedDict as odict

from . import util
from .util import runsyscmd, cacheattr, nested_lookup
from .cmake import CMakeSysInfo


class VisualStudioInfo:
    """encapsulates info on Visual Studio installations"""

    def __init__(self, name):
        cn, toolset = sep_name_toolset(name)
        if cn not in _versions.keys():
            raise Exception("unknown alias: " + name)
        ver = _versions[cn]
        self.name = name
        self.name_without_toolset = cn
        self.toolset = toolset
        self.ver = ver
        self.year = int(re.sub(r'^vs(....).*', r'\1', name))
        self.gen = to_gen(cn)
        self.architecture = parse_architecture(self.gen)
        self.dir = vsdir(ver)
        self.msbuild = msbuild(ver)
        self.devenv = devenv(ver)
        self.vcvarsall = vcvarsall(ver)
        self.is_installed = is_installed(ver)
        self.cxx_compiler = cxx_compiler(ver)
        self.c_compiler = c_compiler(ver)
        if self.toolset is not None:
            self.is_clang = re.search(r'clang', self.toolset) is not None
        else:
            self.is_clang = False


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# vs search order
order = ('vs2015', 'vs2017', 'vs2013', 'vs2012', 'vs2010',
         #'vs2008', 'vs2005',
      )

def find_any():
    for vs in order:
        ver = to_ver(vs)
        if is_installed(vs):
            return VisualStudioInfo(vs)
    return None

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# a reversible dictionary for the VS version numbers
_versions = {
    'vs2015':14, 14:'vs2015', 'vs2015_64':14, 'vs2015_32':14, 'vs2015_arm':14 ,  # nopep8
    'vs2017':15, 15:'vs2017', 'vs2017_64':15, 'vs2017_32':15, 'vs2017_arm':15 ,  # nopep8
    'vs2013':12, 12:'vs2013', 'vs2013_64':12, 'vs2013_32':12, 'vs2013_arm':12 ,  # nopep8
    'vs2012':11, 11:'vs2012', 'vs2012_64':11, 'vs2012_32':11, 'vs2012_arm':11 ,  # nopep8
    'vs2010':10, 10:'vs2010', 'vs2010_64':10, 'vs2010_32':10, 'vs2010_ia64':10,  # nopep8
    'vs2008':9 , 9 :'vs2008', 'vs2008_64':9 , 'vs2008_32':9 , 'vs2008_ia64':9 ,  # nopep8
    'vs2005':8 , 8 :'vs2005', 'vs2005_64':8 , 'vs2005_32':8 ,   # nopep8
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
_sfx = ' Win64' if util.in_64bit() else ''

# a reversible dictionary for the names
_names = {
    'vs2017'      : 'Visual Studio 15 2017' + _sfx , 'Visual Studio 15 2017' + _sfx : 'vs2017'      ,  # nopep8
    'vs2017_32'   : 'Visual Studio 15 2017'        , 'Visual Studio 15 2017'        : 'vs2017_32'   ,  # nopep8
    'vs2017_64'   : 'Visual Studio 15 2017 Win64'  , 'Visual Studio 15 2017 Win64'  : 'vs2017_64'   ,  # nopep8
    'vs2017_arm'  : 'Visual Studio 15 2017 ARM'    , 'Visual Studio 15 2017 ARM'    : 'vs2017_arm'  ,  # nopep8
    'vs2015'      : 'Visual Studio 14 2015' + _sfx , 'Visual Studio 14 2015' + _sfx : 'vs2015'      ,  # nopep8
    'vs2015_32'   : 'Visual Studio 14 2015'        , 'Visual Studio 14 2015'        : 'vs2015_32'   ,  # nopep8
    'vs2015_64'   : 'Visual Studio 14 2015 Win64'  , 'Visual Studio 14 2015 Win64'  : 'vs2015_64'   ,  # nopep8
    'vs2015_arm'  : 'Visual Studio 14 2015 ARM'    , 'Visual Studio 14 2015 ARM'    : 'vs2015_arm'  ,  # nopep8
    'vs2013'      : 'Visual Studio 12 2013' + _sfx , 'Visual Studio 12 2013' + _sfx : 'vs2013'      ,  # nopep8
    'vs2013_32'   : 'Visual Studio 12 2013'        , 'Visual Studio 12 2013'        : 'vs2013_32'   ,  # nopep8
    'vs2013_64'   : 'Visual Studio 12 2013 Win64'  , 'Visual Studio 12 2013 Win64'  : 'vs2013_64'   ,  # nopep8
    'vs2013_arm'  : 'Visual Studio 12 2013 ARM'    , 'Visual Studio 12 2013 ARM'    : 'vs2013_arm'  ,  # nopep8
    'vs2012'      : 'Visual Studio 11 2012' + _sfx , 'Visual Studio 11 2012' + _sfx : 'vs2012'      ,  # nopep8
    'vs2012_32'   : 'Visual Studio 11 2012'        , 'Visual Studio 11 2012'        : 'vs2012_32'   ,  # nopep8
    'vs2012_64'   : 'Visual Studio 11 2012 Win64'  , 'Visual Studio 11 2012 Win64'  : 'vs2012_64'   ,  # nopep8
    'vs2012_arm'  : 'Visual Studio 11 2012 ARM'    , 'Visual Studio 11 2012 ARM'    : 'vs2012_arm'  ,  # nopep8
    'vs2010'      : 'Visual Studio 10 2010' + _sfx , 'Visual Studio 10 2010' + _sfx : 'vs2010'      ,  # nopep8
    'vs2010_32'   : 'Visual Studio 10 2010'        , 'Visual Studio 10 2010'        : 'vs2010_32'   ,  # nopep8
    'vs2010_64'   : 'Visual Studio 10 2010 Win64'  , 'Visual Studio 10 2010 Win64'  : 'vs2010_64'   ,  # nopep8
    'vs2010_ia64' : 'Visual Studio 10 2010 IA64'   , 'Visual Studio 10 2010 IA64'   : 'vs2010_ia64' ,  # nopep8
    'vs2008'      : 'Visual Studio 9 2008' + _sfx  , 'Visual Studio 9 2008' + _sfx  : 'vs2008'      ,  # nopep8
    'vs2008_32'   : 'Visual Studio 9 2008'         , 'Visual Studio 9 2008'         : 'vs2008_32'   ,  # nopep8
    'vs2008_64'   : 'Visual Studio 9 2008 Win64'   , 'Visual Studio 9 2008 Win64'   : 'vs2008_64'   ,  # nopep8
    'vs2008_ia64' : 'Visual Studio 9 2008 IA64'    , 'Visual Studio 9 2008 IA64'    : 'vs2008_ia64' ,  # nopep8
    'vs2005'      : 'Visual Studio 8 2005' + _sfx  , 'Visual Studio 8 2005' + _sfx  : 'vs2005'      ,  # nopep8
    'vs2005_32'   : 'Visual Studio 8 2005'         , 'Visual Studio 8 2005'         : 'vs2005_32'   ,  # nopep8
    'vs2005_64'   : 'Visual Studio 8 2005 Win64'   , 'Visual Studio 8 2005 Win64'   : 'vs2005_64'   ,  # nopep8
}

_architectures = {
    'Visual Studio 15 2017'        : 'x86'    ,
    'Visual Studio 15 2017 Win64'  : 'x86_64' ,
    'Visual Studio 15 2017 ARM'    : 'arm'    ,
    'Visual Studio 14 2015'        : 'x86'    ,
    'Visual Studio 14 2015 Win64'  : 'x86_64' ,
    'Visual Studio 14 2015 ARM'    : 'arm'    ,
    'Visual Studio 12 2013'        : 'x86'    ,
    'Visual Studio 12 2013 Win64'  : '_64'    ,
    'Visual Studio 12 2013 ARM'    : 'arm'    ,
    'Visual Studio 11 2012'        : 'x86'    ,
    'Visual Studio 11 2012 Win64'  : '_64'    ,
    'Visual Studio 11 2012 ARM'    : 'arm'    ,
    'Visual Studio 10 2010'        : 'x86'    ,
    'Visual Studio 10 2010 Win64'  : 'x86_64' ,
    'Visual Studio 10 2010 IA64'   : 'ia64'   ,
    'Visual Studio 9 2008'         : 'x86'    ,
    'Visual Studio 9 2008 Win64'   : 'x86_64' ,
    'Visual Studio 9 2008 IA64'    : 'ia64'   ,
    'Visual Studio 8 2005'         : 'x86'    ,
    'Visual Studio 8 2005 Win64'   : 'x86_64' ,
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def to_name(name_or_gen_or_ver):
    if isinstance(name_or_gen_or_ver, int):
        return _versions[name_or_gen_or_ver]
    else:
        if name_or_gen_or_ver.startswith('vs'):
            return sep_name_toolset(name_or_gen_or_ver)[0]
        n = _names.get(name_or_gen_or_ver)
        if n is not None:
            return n
    raise Exception("could not find '{}'".format(name_or_gen_or_ver))


def to_ver(name_or_gen_or_ver):
    if isinstance(name_or_gen_or_ver, int):
        return name_or_gen_or_ver
    else:
        n = to_name(name_or_gen_or_ver)
        return _versions[n]


def to_gen(name_or_gen_or_ver):
    if isinstance(name_or_gen_or_ver, int):
        name_or_gen_or_ver = _versions[name_or_gen_or_ver]
    if name_or_gen_or_ver.startswith('Visual Studio'):
        return name_or_gen_or_ver
    name_or_gen_or_ver = sep_name_toolset(name_or_gen_or_ver)[0]
    return _names[name_or_gen_or_ver]


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

_toolsets = (
    # vs2017 compiler toolsets
    'v141_clang_c2', 'v141_clang', 'v141_xp', 'v141',
    # vs2015 compiler toolsets
    'v140_clang_c2', 'v140_clang', 'v140_xp', 'v140',
    # vs2013 compiler toolsets
    'v120_xp', 'v120',
    # vs2012 compiler toolsets
    'v110_xp', 'v110',
    # vs2010 compiler toolsets
    'v100_xp', 'v100',
    # vs2008 compiler toolsets
    'v90_xp', 'v90',
    # vs2005 compiler toolsets
    'v80',
    # aliases - implicit compiler toolsets (the same as the chosen VS version)
    'clang_c2', 'clang', 'xp',
)
_toolsets_for_re = sorted(_toolsets, key=lambda x: -len(x))


def sep_name_toolset(name, canonize=True):
    """separate name and toolset"""
    toolset = None
    for t in _toolsets_for_re:
        rx = r'vs.....*_({})$'.format(t)
        if re.search(rx, name):
            toolset = re.sub(rx, r'\1', name)
            name_without_toolset = re.sub(r'(vs.....*)_({})$'.format(t), r'\1', name)
            break
    if toolset is None:
        return name, None
    if toolset not in _toolsets:
        raise Exception("could not parse toolset {} from vs spec {}".format(toolset, name))
    if not canonize:
        return name_without_toolset, toolset
    if toolset in ('clang_c2', 'clang', 'xp'):
        assert re.match('vs....', name)
        year = int(re.sub(r'^vs(....).*', r'\1', name))
        if year == 2017:
            vs_toolset = 'v141_' + toolset
        elif year == 2015:
            vs_toolset = 'v140_' + toolset
        else:
            assert toolset != "clang"
            if year == 2013:
                vs_toolset = 'v120_' + toolset
            elif year == 2012:
                vs_toolset = 'v110_' + toolset
            elif year == 2010:
                vs_toolset = 'v100_' + toolset
            elif year == 2008:
                vs_toolset = 'v90_' + toolset
            elif year == 2005:
                vs_toolset = 'v80_' + toolset
            else:
                raise Exception("toolset not implemented for " + name + ". toolset="+toolset)
    else:
        vs_toolset = toolset
    if vs_toolset.endswith('clang'):
        vs_toolset += '_c2'
    return name_without_toolset, vs_toolset


def parse_toolset(name, canonize=True):
    _, toolset = sep_name_toolset(name, canonize)
    return toolset


def parse_architecture(name_or_gen_or_ver):
    gen = to_gen(name_or_gen_or_ver)
    a = _architectures[gen]
    return a


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def vsdir(name_or_gen_or_ver):
    """get the directory where VS is installed"""
    ver = to_ver(name_or_gen_or_ver)
    d = ""
    if ver < 15:
        progfilesx86 = os.environ['ProgramFiles(x86)']
        d = os.path.join(progfilesx86, 'Microsoft Visual Studio ' + str(ver) + '.0')
        if not os.path.exists(d):
            try:
                v = os.environ['VS{}0COMNTOOLS'.format(str(ver))]
                d = os.path.abspath(os.path.join(v, '..', '..'))
            except:
                pass
    elif ver == 15:
        # VS 2017+ is no longer a singleton, and may be installed anywhere;
        # also, the environment variable VS***COMNTOOLS no longer exists.
        def fn():
            try:
                idata = _vs2017_get_instance_data()
                path = nested_lookup(idata, 'installationPath')
                return path
            except:
                return ""
        d = cacheattr(sys.modules[__name__], '_vs2017dir', fn)
    else:
        raise Exception('VS Version not implemented: ' + str(ver))
    return d


def devenv(name_or_gen_or_ver):
    """get the path to vcvarsall.bat"""
    ver = to_ver(name_or_gen_or_ver)
    d = vsdir(ver)
    # devenv can have different names:
    # see http://stackoverflow.com/questions/7818543/no-devenv-file-in-microsoft-visual-express-10
    for n in ('devenv.exe', 'WDExpress.exe', 'VSWinExpress.exe'):
        s = os.path.join(d, 'Common7', 'IDE', n)
        if os.path.exists(s):
            return s
    return None


def vcvarsall(name_or_gen_or_ver):
    """get the path to vcvarsall.bat"""
    ver = to_ver(name_or_gen_or_ver)
    d = vsdir(ver)
    if ver < 15:
        s = os.path.join(d, 'VC', 'vcvarsall.bat')
    elif ver == 15:
        s = os.path.join(d, 'VC', 'Auxiliary', 'Build', 'vcvarsall.bat')
    else:
        raise Exception('VS Version not implemented: ' + str(ver))
    return s


def msbuild(name_or_gen_or_ver):
    """get the MSBuild.exe path"""
    ver = to_ver(name_or_gen_or_ver)
    if ver < 12:   # VS2012 and below
        import winreg as wr
        msbuild = None
        msbvers = ('4.0', '3.5', '2.0')
        for v in msbvers:
            key = "SOFTWARE\\Microsoft\\MSBuild\\ToolsVersions\\{}\\MS‌​BuildToolsPath"
            try:
                val = wr.OpenKey(wr.HKEY_LOCAL_MACHINE, key.format(ver), 0, wr.KEY_READ)
                msbuild = os.path.join(val, 'MSBuild.exe')
                break
            except:
                pass
        # default to some probable value if no registry key was found
        if msbuild is None:
            val = 'C:\\Windows\Microsoft.NET\Framework{}\\v{}\\MSBuild.exe'
            for v in msbvers:
                msbuild = val.format('64' if util.in_64bit() else '', '3.5')
                if os.path.exists(msbuild):
                    break
    else:
        root = os.environ['ProgramFiles(x86)'] if ver < 15 else vsdir(ver)
        val = '{}\\MSBuild\\{}.0\\bin\\{}MSBuild.exe'
        msbuild = val.format(root, ver, 'amd64\\' if util.in_64bit() else '')
    return msbuild


def cxx_compiler(name_or_gen_or_ver):
    if not is_installed(name_or_gen_or_ver):
        return ""
    return CMakeSysInfo.cxx_compiler(to_gen(name_or_gen_or_ver))


def c_compiler(name_or_gen_or_ver):
    if not is_installed(name_or_gen_or_ver):
        return ""
    return CMakeSysInfo.c_compiler(to_gen(name_or_gen_or_ver))


def is_installed(name_or_gen_or_ver):
    ver = to_ver(name_or_gen_or_ver)
    return cacheattr(sys.modules[__name__], '_is_installed_'+str(ver), lambda: _is_installed_impl(ver))


def _is_installed_impl(ver):
    assert isinstance(ver, int)
    if ver < 15:
        import winreg as wr
        key = "SOFTWARE\\Microsoft\\VisualStudio\\{}.0"
        try:
            wr.OpenKey(wr.HKEY_LOCAL_MACHINE, key.format(ver), 0, wr.KEY_READ)
            # fail if we can't find the dir
            if not os.path.exists(vsdir(ver)):
                return False
            # apparently the dir is not enough, so check also vcvarsall
            if not os.path.exists(vcvarsall(ver)):
                return False
            return True
        except:
            return False
    else:
        #
        # ~~~~~~~~~~~~~~ this is fragile.... ~~~~~~~~~~~~~~
        #
        # Unlike earlier versions, VS2017 is no longer a singleton installation.
        # Each VS2017 installed instance keeps a store of its data under
        # %ProgramData%\Microsoft\VisualStudio\Packages\_Instances\<hash>\state.json
        #
        # this info was taken from:
        # http://stackoverflow.com/questions/40694598/how-do-i-call-visual-studio-2017-rcs-version-of-msbuild-from-a-bat-files
        for i in iter(_vs2017_get_instances().keys()):
            try:
                d = _vs2017_get_instance_data(i)
            except:
                continue
            # check that the version matches
            version_string = nested_lookup(d, 'catalogInfo', 'buildVersion')
            version_number = int(re.sub(r'(\d\d).*', r'\1', version_string))
            if version_number != ver:
                continue
            # check that the directory exists
            install_dir = nested_lookup(d, 'installationPath')
            if not os.path.exists(install_dir):
                continue
            # maybe further checks are necessary?
            # For now we stop here, and accept that this installation exists.
            return True
        return False


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def _vs2017_get_instance_data(which_instance=None):
    i = _vs2017_resolve_instance(which_instance)
    if i is None and which_instance is not None:
        err = "could not find a vs2017 instance named {}"
        raise Exception(err.format(which_instance))
    def fn():
        instances = _vs2017_get_instances()
        with open(instances[i], encoding="utf8") as json_str:
            d = json.load(json_str)
            return d
    return cacheattr(sys.modules[__name__], "_vs2017_instance_data_" + i, fn)


def _vs2017_resolve_instance(which_instance=None):
    progdata = os.environ['ProgramData']
    d = os.path.join(progdata, 'Microsoft', 'VisualStudio', 'Packages', '_Instances')
    instances = _vs2017_get_instances()
    if instances is None:
        raise Exception("could not find a vs2017 instance in " + d)
    i = None
    for j in iter(instances.keys()):
        if (which_instance is None) or (which_instance == j):
            i = j
            break
    return i


def _vs2017_get_instances():
    def fn():
        d = odict()
        progdata = os.environ['ProgramData']
        instances_dir = os.path.join(progdata, 'Microsoft', 'VisualStudio', 'Packages', '_Instances')
        if not os.path.exists(instances_dir):
            return d
        pat = os.path.join(instances_dir, '*', 'state.json')
        instances = glob.glob(pat)
        if not instances:
            return d
        for i in instances:
            d[os.path.basename(os.path.dirname(i))] = i
        return d
    return cacheattr(sys.modules[__name__], "_vs2017_instances", fn)



from collections import OrderedDict as odict
from ruamel import yaml
import copy

from .named_item import NamedItem
from . import util


def get_name_for_flags(compiler):
    if isinstance(compiler, str):
        return compiler
    sn = compiler.name_for_flags
    if compiler.is_msvc:
        if compiler.vs.is_clang:
            sn = 'clang'
        else:
            sn = 'vs'
    return sn


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class FlagAliases:

    def __init__(self, **kwargs):
        if kwargs.get('txt') is not None:
            raise Exception("not implemented")
            # self.compilers, self.flags = load_txt(kwargs.get('txt'))
        elif kwargs.get('yml') is not None:
            self.compilers, self.flags = load_yml(kwargs.get('yml'))
        else:
            self.flags = odict(**kwargs)
            self.compilers = get_all_compilers(self.flags)

    def merge_from(self, other):
        self.flags = merge(self.flags, other.flags)
        self.compilers = get_all_compilers(self.flags)

    def get(self, name, compiler=None):
        opt = self.flags.get(name)
        if opt is None:
            raise Exception("could not find flag alias: " + name)
        if compiler is not None:
            return opt.get(compiler)
        return opt

    def as_flags(self, spec, compiler=None):
        out = []
        for s in spec:
            if isinstance(s, CFlag):
                out.append(s)
            else:
                f = self.flags.get(s)
                if f is not None:
                    out.append(f)
                else:
                    ft = CFlag(name=s, desc=s)
                    ft.set(compiler, s)
                    out.append(ft)
        if compiler is not None:
            out = [f.get(compiler) for f in out]
        return out

    def as_defines(self, spec, compiler=None):
        out = []
        wf = '/D' if get_name_for_flags(compiler) == 'vs' else '-D'
        prev = None
        for s in spec:
            if prev != wf and not s.startswith(wf):
                out.append(wf)
            out.append(s)
            prev = s
        return out


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
class CFlag(NamedItem):

    def __init__(self, name, desc='', **kwargs):
        super().__init__(name)
        self.desc = desc
        self.compilers = []
        for k, v in kwargs.items():
            self.set(k, v)

    def get(self, compiler):
        compseq = (compiler, 'gcc', 'g++', 'vs')  # not really sure about this
        s = None
        for c in compseq:
            sn = get_name_for_flags(c)
            if hasattr(self, sn):
                s = getattr(self, sn)
                break
        if s is None:
            util.logwarn('compiler not found: ', compiler, self.__dict__)
            s = ''
        # print(self, sn, s)
        return s

    def set(self, compiler, val=''):
        sn = get_name_for_flags(compiler)
        setattr(self, sn, val)
        if sn not in self.compilers:
            self.compilers.append(sn)

    def add_compiler(self, compiler):
        sn = get_name_for_flags(compiler)
        if not hasattr(self, sn):
            self.set(sn)
        if sn not in self.compilers:
            self.compilers.append(sn)

    def merge_from(self, that):
        for k, v in that.__dict__.items():
            if not __class__.is_compiler_name(k):
                continue
            v = that.get(k)
            self.set(k, v)

    @staticmethod
    def is_compiler_name(s):
        return not (s.startswith('__') or s == 'name' or s == 'desc'
                    or s == 'compilers')


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def dump_yml(comps, flags):
    """dump the given compilers and flags pair into a yml string"""
    txt = ""
    for n, f in flags.items():
        txt += n + ':\n'
        if f.desc:
            txt += '    desc: ' + f.desc + '\n'
        # join equal flag values into comma-separated keys:
        # eg compiler1,compiler2,compiler3: flag value
        rd = odict()
        dr = odict()
        done = odict()
        # store reverse information: flag: compiler1,compiler2,compiler3
        for comp in comps:
            v = getattr(f, comp)
            if not rd.get(v):
                rd[v] = ''
            else:
                rd[v] += ','
            rd[v] += comp
            dr[comp] = v
        # make sure compilers are not repeated
        for comp in comps:
            done[comp] = False
        # now lookup, write and mark
        for comp in comps:
            if done[comp]:
                continue
            val = dr[comp]
            key = rd[val]
            if val:
                txt += '    ' + key + ': ' + val + '\n'
            for ccomp in key.split(','):
                done[ccomp] = True
    return txt


def load_txt(yml_txt):
    """load a yml txt into a compilers, flags pair"""
    YAML = yaml.YAML()
    dump = YAML.load(yml_txt)
    fa = dump.get('flag_aliases', dump)
    return load_yml(fa)


def load_yml(dump):
    fd = odict(dump)
    # gather the list of compilers
    comps = []
    for n, yf in fd.items():
        for c in yf:
            c = str(c)
            if CFlag.is_compiler_name(c):
                for cc in c.split(','):
                    if cc not in comps:
                        comps.append(cc)
    # now load flags, making sure that all have the same compiler names
    flags = odict()
    for n, yf in fd.items():
        f = CFlag(n)
        for c in comps:
            f.add_compiler(c)
        for comp_, val in yf.items():
            for comp in comp_.split(','):
                if comp == 'desc':
                    f.desc = val
                else:
                    f.set(comp, val)
                #print("load yml: flag value for", comp, ":", val, "--------->", f.get(comp))
        flags[n] = f
    return comps, flags


def merge(flags, into_flags=None):
    """merge flags into the previously existing flags"""
    into_flags = into_flags if into_flags is not None else known_flags
    result_flags = into_flags
    comps = get_all_compilers(flags, into_flags)
    result_flags = copy.deepcopy(into_flags)
    for k, v in flags.items():
        if k in result_flags:
            result_flags[k].merge_from(v)
        else:
            result_flags[k] = copy.deepcopy(v)
    for f in result_flags:
        for c in comps:
            result_flags[k].add_compiler(c)
    return result_flags


def get_all_compilers(*flag_dicts):
    comps = []
    for f in flag_dicts:
        for k, v in f.items():
            for c in v.compilers:
                if c not in comps:
                    comps.append(c)
    return comps



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os.path
import sys
import site
import re
import glob


if sys.version_info < (3, 3):
    # this is because of subprocess. That code is in c4/cmany/util.py.
    msg = 'cmany requires at least Python 3.3. Current version is {}. Sorry.'
    sys.exit(msg.format(sys.version_info))


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


def read_manifest():
    thisd = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(thisd, 'MANIFEST.in')) as f:
        lines = [re.sub(r'include (.*)$', r'\1', l) for l in f.readlines()]
        lines = [os.path.join(thisd, l) for l in lines]
        return lines


def readreqs(*rnames):
    def _skipcomment(line):
        return line if (line and not line.startswith('--')
                        and not line.startswith('#')) else ""
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        l = [_skipcomment(line.strip()) for line in f]
    return l


def get_binaries_directory():
    """Return the installation directory, or None
    http://stackoverflow.com/questions/36187264"""
    if '--user' in sys.argv:
        paths = (site.getusersitepackages(),)
    else:
        if hasattr(site, 'getsitepackages'):
            print("setup: site-packages", site.getsitepackages())
            print("setup: site-user-packages", site.getusersitepackages())
            print("setup: site-prefix", os.path.dirname(sys.executable))
            paths = site.getsitepackages()
        else:
            print("setup: no site.getsitepackages()...")
            py_prefix = os.path.dirname(sys.executable)
            paths = [
                py_prefix + '/lib/site-packages',
                py_prefix + '/lib/site-packages',
                py_prefix + '/lib',
            ]
        py_version = '{}.{}'.format(sys.version_info[0], sys.version_info[1])
        paths += (s.format(py_version) for s in (
            sys.prefix + '/lib/python{}/site-packages/',
            sys.prefix + '/lib/python{}/dist-packages/',
            sys.prefix + '/local/lib/python{}/site-packages/',
            sys.prefix + '/local/lib/python{}/dist-packages/',
            '/Library/Python/{}/site-packages/',
        ))
    for path in paths:
        if os.path.exists(path):
            print('setup: installation path:', path)
            return path
    raise Exception('setup: no installation path found', file=sys.stderr)
    return None


def get_data_files():
    dest = get_binaries_directory()
    d = lambda d: os.path.join(dest, d)
    return [
          (d("c4/cmany/"), [
              "LICENSE.txt",
              "README.rst",
              "requirements.txt",
              "requirements_test.txt"
          ]),
          (d("c4/cmany/conf"), [
              "conf/cmany.yml"
          ]),
          (d("c4/cmany/doc"),
              glob.glob("doc/_build/text/*.txt")),
      ]


setup(name="cmany",
      version="0.1.0.post5",
      description="CMake build tree batching tool",
      long_description=read('README.rst') + "\n" + read('LICENSE.txt'),
      url="https://github.com/biojppm/cmany",
      download_url="https://github.com/biojppm/cmany",
      license="License :: OSI Approved :: MIT License",
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Intended Audience :: Developers",
          "Development Status :: 3 - Alpha",
          "Programming Language :: Python :: 3",
          "Programming Language :: C",
          "Programming Language :: C++",
          "Topic :: Software Development :: Build Tools",
          "Topic :: Software Development :: Compilers",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: Utilities",
      ],
      keywords=["cmake", "c++", "c"],
      author="Joao Paulo Magalhaes",
      author_email="dev@jpmag.me",
      zip_safe=False,
      namespace_packages=['c4'],
      packages=find_packages('src', exclude=['test']),
      package_dir={'': 'src'},
      entry_points={'console_scripts': ['cmany=c4.cmany.main:cmany_main'], },
      install_requires=readreqs('requirements.txt'),
      tests_require=readreqs('requirements_test.txt'),
      include_package_data=True,
      # package_data={'c4.cmany':read_manifest()},
      data_files=get_data_files()
)



import re

from .build_item import BuildItem
from . import util


# -----------------------------------------------------------------------------
class Architecture(BuildItem):
    """Specifies a processor architecture"""

    @staticmethod
    def default():
        """return the architecture of the current machine"""
        return Architecture(__class__.default_str())

    @staticmethod
    def default_str():
        # s = CMakeSysInfo.architecture()
        # if s == "amd64":
        #     s = "x86_64"
        # return s
        if util.in_64bit():
            return "x86_64"
        elif util.in_32bit():
            return "x86"

    @property
    def is64(self):
        def fn():
            s = re.search('64', self.name)
            return s is not None
        return util.cacheattr(self, "_is64", fn)

    @property
    def is32(self):
        return not self.is64 and not self.is_arm

    @property
    def is_arm(self):
        return "arm" in self.name.lower()



#__import__('pkg_resources').declare_namespace(__name__)

from .conf import *

from .build_flags import BuildFlags as BuildFlags
from .system import System as System
from .architecture import Architecture as Architecture
from .compiler import Compiler as Compiler
from .build_type import BuildType as BuildType
from .variant import Variant as Variant
from .build import Build as Build
from .project import Project as Project



from .named_item import NamedItem as NamedItem
from . import flags as c4flags

from . import util

# -----------------------------------------------------------------------------
class BuildFlags(NamedItem):

    attrs = ('cmake_vars', 'defines', 'cflags', 'cxxflags', 'toolchain')

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.cmake_vars = kwargs.get('cmake_vars', [])
        self.defines = kwargs.get('defines', [])
        self.cflags = kwargs.get('cflags', [])
        self.cxxflags = kwargs.get('cxxflags', [])
        self.toolchain = kwargs.get('toolchain')
        # self.include_dirs = kwargs['include_dirs']
        # self.link_dirs = kwargs['link_dirs']

    def empty(self):
        if self.cmake_vars:
            return False
        if self.defines:
            return False
        if self.cflags:
            return False
        if self.cxxflags:
            return False
        if self.toolchain:
            return False
        return True

    def resolve_flag_aliases(self, compiler, aliases):
        self.defines = aliases.as_defines(self.defines, compiler)
        self.cflags = aliases.as_flags(self.cflags, compiler)
        self.cxxflags = aliases.as_flags(self.cxxflags, compiler)

    def append_flags(self, other, append_to_name=True):
        """other will take precedence, ie, their options will come last"""
        if append_to_name and other.name:
            self.name += '_' + other.name
        self.cmake_vars += other.cmake_vars
        self.defines += other.defines
        self.cflags += other.cflags
        self.cxxflags += other.cxxflags
        # self.include_dirs += other.include_dirs
        # self.link_dirs += other.link_dirs
        self.toolchain = __class__.merge_toolchains(self.toolchain, other.toolchain)

    def log(self, log_fn=print, msg=""):
        t = "BuildFlags[{}]: {}".format(self.name, msg)
        log_fn(t, "cmake_vars=", self.cmake_vars)
        log_fn(t, "defines=", self.defines)
        log_fn(t, "cxxflags=", self.cxxflags)
        log_fn(t, "cflags=", self.cflags)
        log_fn(t, "toolchain=", self.toolchain)

    @staticmethod
    def merge_toolchains(tc1, tc2):
        if ((tc1 != tc2) and (tc1 is not None and tc2 is not None)):
            raise Exception("conflicting toolchains: " + tc1 + " vs. " + tc2)
        if tc1 is None and tc2 is not None:
            tc1 = tc2
        return tc1

    def save_config(self, yml_node):
        for n in __class__.attrs:
            a = getattr(self, n)
            if a:
                a = __class__.flag_list_to_str(a)
                yml_node[n] = a

    def load_config(self, yml_node):
        for n in __class__.attrs:
            a = yml_node.get(n)
            if a is not None:
                if n != 'toolchain':
                    a = __class__.flag_str_to_list(a)
                setattr(self, n, a)

    @staticmethod
    def flag_list_to_str(li):
        if isinstance(li, str):
            return li
        return " ".join(li)

    @staticmethod
    def flag_str_to_list(s):
        if isinstance(s, list):
            return s
        return util.splitesc_quoted(s, ' ')




from ruamel import yaml
import os.path
from collections import OrderedDict as odict

from . import util


class Conan:

    def __init__(self):
        util.cacheattr(Conan, 'settings', Conan.load_settings)

    def install(self, build):
        with util.setcwd(build.builddir, silent=False):
            cmd = (['conan', 'install', '--build=missing'] +
                   self.translate_os(build.system) +
                   self.translate_architecture(build.architecture) +
                   self.translate_compiler(build.compiler) +
                   self.translate_buildtype(build.build_type) +
                   [os.path.abspath(build.projdir)])
            util.runsyscmd(cmd)

    @staticmethod
    def load_settings():
        conandir = os.path.expanduser("~/.conan/")
        if not os.path.exists(conandir):
            return
        settings_file = os.path.join(conandir, 'settings.yml')
        with open(settings_file) as f:
            txt = f.read()
            YAML = yaml.YAML()
            data = YAML.load(txt)
            settings = odict(data)
            return settings

    def translate_os(self, system):
        s = system.name
        if s == 'windows':
            s = 'Windows'
        conan = Conan.settings['os']
        if s in conan:
            return ['-s', 'os=' + s]
        msg = "system not found in conan: {}. Must be one of {}"
        raise Exception(msg.format(s, conan))

    def translate_architecture(self, architecture):
        s = architecture.name
        conan = Conan.settings['arch']
        if s in conan:
            return ['-s', 'arch=' + s]
        msg = "architecture not found in conan: {}. Must be one of {}"
        raise Exception(msg.format(s, conan))

    def translate_compiler(self, compiler):
        s = compiler.name
        if compiler.is_msvc:
            return ['-s', 'compiler=Visual Studio',
                    '-s', 'compiler.version='+str(compiler.vs.ver)]
        elif compiler.shortname == 'gcc':
            libcxx = 'libstdc++11'  # FIXME
            return ['-s', 'compiler=gcc',
                    '-s', 'compiler.version=' + str(compiler.version),
                    '-s', 'compiler.libcxx=' + str(libcxx)]
        conan = Conan.settings['compiler']
        if s in conan:
            return ['-s', 'compiler=' + s]
        msg = "compiler not found in conan: {}. Must be one of {}"
        raise Exception(msg.format(s, conan))

    def translate_build_type(self, build_type):
        s = build_type.name
        conan = Conan.settings['build_type']
        if s in conan:
            return ['-s', 'build_type=' + s]
        msg = "build type not found in conan: {}. Must be one of {}"
        raise Exception(msg.format(s, conan))

    def translate_variant(self, variant):
        return []
